@@ -2,4 +2,7 @@ import pkg_resources
 import pandas as pd
 
 titanic_f = pkg_resources.resource_filename(__name__, 'titanic/titanic.csv')
-titanic_df = pd.read_csv(titanic_f)
\ No newline at end of file
+titanic_df = pd.read_csv(titanic_f)
+
+ibd_f = pkg_resources.resource_filename(__name__, 'ibd/IBD.csv')
+ibd_df = pd.read_csv(titanic_f)
\ No newline at end of file
