@@ -12,10 +12,15 @@ from sklearn.linear_model import LogisticRegression
 from sklearn.preprocessing import LabelEncoder
 
 class LogisticRegressionClass:
-    def __init__(self,df,response,sig_level=0.05,max_iter=500):
+    def __init__(self,df,response,sig_level=0.05,max_iter=500,cols_to_keep_static=[],cols_to_try_individually=[]):
         '''
         :param df: a dataframe
         :param response: a string. This must be an existing column in df
+        :param sig_level: a float. The significance level the forward selection will use
+        :param max_iter: an integer. The maximum iterations the solvers will use to try to converge
+        :param cols_to_keep_static: a list. Used in forward selection to not omit these columns
+        :param cols_to_try_individually: a list. The columns to test in a regression one at a time to identify which
+            one has the greatest relationship with the response controlled for the cols_to_keep_static
         '''
 
         # attach attributes to the object
@@ -23,6 +28,10 @@ class LogisticRegressionClass:
         self.response = response
         self.sig_level = sig_level
         self.max_iter=max_iter
+        self.warnings = ''
+        self.error_message = ''
+        self.cols_to_keep_static = cols_to_keep_static
+        self.cols_to_try_individually = cols_to_try_individually
 
     def prepare_data(self,df,response):
         y = df[response]
@@ -33,10 +42,12 @@ class LogisticRegressionClass:
 
     def logistic_regression_utility_check_response(self,series):
         if (len(series.unique()) > 2):
+            self.error_message = 'The response variable has more than 2 categories and is not suitable for logistic regression'
             print('The response variable has more than 2 categories and is not suitable for logistic regression')
             return False
 
         if (not is_numeric_dtype(series)):
+            self.error_message = self.error_message + '\n' + 'The response variable should be binary 0 and 1 and numeric type (i.e. int)'
             print('The response variable should be binary 0 and 1 and numeric type (i.e. int)')
             return False
 
@@ -94,18 +105,40 @@ class LogisticRegressionClass:
 
     def prepare_categories(self,df, response, drop=False):
         cat_cols = list(filter(lambda x: not is_numeric_dtype(df[x]), df.columns))
-        cat_cols = list(set(cat_cols) - {response})
+        cat_cols = list(set(cat_cols) - {response} - set(self.cols_to_keep_static))
         df = pd.get_dummies(df, columns=cat_cols, drop_first=drop)
+        df = pd.get_dummies(df, columns=self.cols_to_keep_static, drop_first=True)
+
+        self.cols_to_keep_static_dummified = []
+        for col in self.cols_to_keep_static:
+            for col_dummy in df.columns:
+                if col in col_dummy:
+                    self.cols_to_keep_static_dummified.append(col_dummy)
 
         return df
 
-    def get_interpretation(self,result,feature_list):
+    def get_interpretation(self,result=None,feature_list=None,df=None):
         '''
         Given a trained model, calculate the average probabilities due to feature changes
         '''
 
+        if (result is None) or (feature_list is None):
+            try:
+                feature_list = self.X_with_feature_selection.columns
+                result = self.result_with_feature_selection
+            except:
+                feature_list = self.X.columns
+                try:
+                    result = self.result
+                except:
+                    result = self.basic_result
+
+
         # take a copy of the original df and prepare the dataset
-        df_temp = self.df.copy()
+        if df is None:
+            df = self.df.copy()
+
+        df_temp = df.copy()
         df_temp = self.prepare_categories(df_temp, self.response, drop=False)
         X, y = self.prepare_data(df_temp,self.response)
 
@@ -113,19 +146,45 @@ class LogisticRegressionClass:
         if 'const' not in full_feature_list:
             full_feature_list = ['const'] + full_feature_list
 
+        # comparative uplift section
+        comparative_dict = dict()
+        for col1 in df.columns:
+            for col2 in full_feature_list:
+                # if this feature was dummified
+                if col1 + '_' in col2:
+                    t = X[full_feature_list].copy()
+                    # First get prediction with 0
+                    t[col2] = 0
+                    comparative_dict[col2] = [result.predict(t).mean()]
+                    # Then get prediction with 1
+                    t[col2] = 1
+                    comparative_dict[col2].append(result.predict(t).mean())
+                elif col1 == col2:
+                    t = X[full_feature_list].copy()
+                    # first get prediction with average
+                    t[col2] = t[col2].mean()
+                    comparative_dict[col2] = [result.predict(t).mean()]
+                    # then get prediction with +1
+                    t[col2] = t[col2] + 1
+                    comparative_dict[col2].append(result.predict(t).mean())
+
+        feature_interpretability_comparative_df = pd.DataFrame(comparative_dict).T
+        feature_interpretability_comparative_df.columns = ['Prediction_average_or_without','Prediction_add1_or_with']
+        self.feature_interpretability_comparative_df = feature_interpretability_comparative_df
+
         # get a base probability (this is just the average probability)
         base_probability = result.predict(X[full_feature_list]).mean()
         probability_dict = dict()
         probability_dict['base'] = base_probability
 
         # for each column in the original df
-        for col in self.df.columns:
+        for col in df.columns:
             # for each column in the result's feature list
             for col2 in feature_list:
                 # check if this feature was dummified from this column
                 if col + '_' in col2:
                     # if this feature was dummified from this column then update this column to be this feature value
-                    df_temp = self.df.copy()
+                    df_temp = df.copy()
                     df_temp[col] = col2.replace(col + '_', '')
                     df_temp = self.prepare_categories(df_temp, self.response, drop=False)
                     X, y = self.prepare_data(df_temp, self.response)
@@ -140,7 +199,7 @@ class LogisticRegressionClass:
                     probability_dict[col2] = probability
                 elif col == col2:
                     # if this column was not dummified then it is numeric so add 1 to it
-                    df_temp = self.df.copy()
+                    df_temp = df.copy()
                     df_temp[col] = df_temp[col] + 1
                     df_temp = self.prepare_categories(df_temp, self.response, drop=False)
                     X, y = self.prepare_data(df_temp, self.response)
@@ -175,19 +234,32 @@ class LogisticRegressionClass:
         return result
 
 
-    def log_reg(self):
-        if not self.logistic_regression_utility_check_response(self.df[self.response]):
+    def log_reg(self,df=None):
+        if df is None:
+            df1 = self.df[~self.df.isna().any(axis=1)].copy()
+
+            if len(df1) < len(self.df):
+                warning_message = 'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(len(self.df),
+                                                                                                               len(df1))
+                warnings.warn(warning_message)
+                print(warning_message)
+
+                self.warnings = self.warnings + '\n' + warning_message
+        else:
+            df1 = df[~df.isna().any(axis=1)].copy()
+
+            if len(df1) < len(df):
+                warning_message = 'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(
+                    len(df),
+                    len(df1))
+                warnings.warn(warning_message)
+                print(warning_message)
+
+                self.warnings = self.warnings + '\n' + warning_message
+
+        if not self.logistic_regression_utility_check_response(df1[self.response]):
             return None
 
-        df1 = self.df[~self.df.isna().any(axis=1)].copy()
-
-        if len(df1) < len(self.df):
-            warnings.warn(
-                'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(len(self.df),
-                                                                                                           len(df1)))
-            print('There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(len(self.df),
-                                                                                                           len(df1)))
-
         df1 = self.prepare_categories(df1,self.response,drop=True)
 
         result = self.log_reg_basic(df1)
@@ -197,21 +269,36 @@ class LogisticRegressionClass:
 
         return result
 
-    def log_reg_with_feature_selection(self,run_for=0,verbose=True):
+    def log_reg_with_feature_selection(self,df=None,run_for=0,verbose=True):
         # start the timer in case the is a time limit specified
         start_time = time.time()
 
-        # get rid of nans. There should be no nans. Imputation should be performed prior to this point
-        df1 = self.df[~self.df.isna().any(axis=1)].copy()
+        if df is None:
+            # get rid of nans. There should be no nans. Imputation should be performed prior to this point
+            df1 = self.df[~self.df.isna().any(axis=1)].copy()
 
-        # show a warning to let the user know of the droppped nans
-        if len(df1) < len(self.df):
-            warnings.warn(
-                'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(len(self.df),
-                                                                                                           len(df1)))
-            print(
-                'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(len(self.df),
-                                                                                                           len(df1)))
+            # show a warning to let the user know of the droppped nans
+            if len(df1) < len(self.df):
+                warning_message = 'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(
+                    len(self.df),
+                    len(df1))
+                warnings.warn(warning_message)
+                print(warning_message)
+
+                self.warnings = self.warnings + '\n' + warning_message
+        else:
+            # get rid of nans. There should be no nans. Imputation should be performed prior to this point
+            df1 = df[~df.isna().any(axis=1)].copy()
+
+            # show a warning to let the user know of the droppped nans
+            if len(df1) < len(df):
+                warning_message = 'There are NaNs in the dataset. After removing NaNs, the rows reduce from {} to {}'.format(
+                    len(df),
+                    len(df1))
+                warnings.warn(warning_message)
+                print(warning_message)
+
+                self.warnings = self.warnings + '\n' + warning_message
 
         # check that the response is in the correct format to perform logistic regression
         if not self.logistic_regression_utility_check_response(df1[self.response]):
@@ -231,10 +318,10 @@ class LogisticRegressionClass:
                     len(df1.columns), len(df1)))
 
         # the initial list of features
-        remaining = list(set(df1.columns) - {self.response})
+        remaining = list(set(df1.columns) - {self.response} - set(self.cols_to_keep_static_dummified))
 
         # this holds the tried and successful feature set
-        full_feature_set = []
+        full_feature_set = self.cols_to_keep_static_dummified
 
         # get the first logistic regression output for only the constant/base model
         first_result = self.log_reg_basic(df1[[self.response]])
@@ -269,7 +356,7 @@ class LogisticRegressionClass:
                 this_feature_set = full_feature_set + [col]
                 try:
                     result = self.log_reg_basic(df1[this_feature_set + [self.response]])
-                except:
+                except Exception as e:
                     remaining.remove(col)
                     continue
 
@@ -321,6 +408,38 @@ class LogisticRegressionClass:
 
         return final_result
 
+    def log_reg_one_at_a_time(self,with_feature_selection=False,get_interpretability=False):
+
+        dic = dict()
+        df1 = self.df.copy()
+        df1 = df1[[self.response]+self.cols_to_keep_static + self.cols_to_try_individually].copy()
+
+        for this_col_to_try in self.cols_to_try_individually:
+            if with_feature_selection:
+                result = self.log_reg_with_feature_selection(df=df1[self.cols_to_keep_static + [self.response, this_col_to_try]])
+                if get_interpretability:
+                    self.get_interpretation(self.result_with_feature_selection,self.final_feature_set
+                                            ,df=df1[self.cols_to_keep_static + [self.response, this_col_to_try]])
+            else:
+                result = self.log_reg(df=df1[self.cols_to_keep_static + [self.response,this_col_to_try]])
+                if get_interpretability:
+                    self.get_interpretation(self.result, self.X.columns
+                                            , df=df1[self.cols_to_keep_static + [self.response, this_col_to_try]])
+            for col in list(filter(lambda x: this_col_to_try in x,result.params.index)):
+                if get_interpretability:
+                    dic[col] = [result.params[col],result.pvalues[col],self.feature_interpretability_df['Probability'][col],
+                                self.feature_interpretability_df['Probability']['base']]
+                else:
+                    dic[col] = [result.params[col], result.pvalues[col]]
+
+        df_one_at_a_time = pd.DataFrame(dic).T
+        if get_interpretability:
+            df_one_at_a_time.columns = ['Coefficient','Pvalue','Controlled Probability','Base Probability']
+        else:
+            df_one_at_a_time.columns = ['Coefficient','Pvalue']
+        self.df_one_at_a_time = df_one_at_a_time
+        return df_one_at_a_time
+
 
 def unit_test_1():
     print('Unit test 1...')
@@ -370,7 +489,16 @@ auc = 0.85'''
     assert (result_required == result_actual)
 
     result_required = [0.4061624649859944, 0.24581360407372246, 0.795820946281563, 0.3999162261394402, 0.3539768140703711, 0.39737068898845873, 0.4064703482674913]
-    result_actual = list(my_logistic_regresion_class.get_interpretation(my_logistic_regresion_class.basic_result,my_logistic_regresion_class.X.columns)['Probability'])
+    result_actual = list(my_logistic_regresion_class.get_interpretation()['Probability'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (result_required == result_actual)
+
+    result_required = [0.22395117322839098, 0.3558311002657131, 0.3977261128754439, 0.4027792667251947,
+                       0.4068054774641371, 0.8543626808843827]
+    result_actual = sorted(list(my_logistic_regresion_class.feature_interpretability_comparative_df['Prediction_add1_or_with']))
 
     result_required = list(map(lambda x: round(x, 2), result_required))
     result_actual = list(map(lambda x: round(x, 2), result_actual))
@@ -430,9 +558,7 @@ auc = 0.85'''
 
     result_required = [0.4061624649859944, 0.24581360407372246, 0.795820946281563, 0.3999162261394402,
                        0.3539768140703711, 0.39737068898845873, 0.4064703482674913]
-    result_actual = list(my_logistic_regresion_class.get_interpretation(my_logistic_regresion_class.basic_result,
-                                                                        my_logistic_regresion_class.X.columns)[
-                             'Probability'])
+    result_actual = list(my_logistic_regresion_class.get_interpretation()['Probability'])
 
     result_required = list(map(lambda x: round(x, 2), result_required))
     result_actual = list(map(lambda x: round(x, 2), result_actual))
@@ -489,9 +615,16 @@ auc = 0.85'''
     assert (result_required == result_actual)
 
     result_required = [0.40616246498599445, 0.24581360407372244, 0.23089275089703268, 0.3999162261394402, 0.3539768140703711, 0.39737068898845873, 0.4064703482674913]
-    result_actual = list(my_logistic_regresion_class.get_interpretation(my_logistic_regresion_class.result,
-                                                                        my_logistic_regresion_class.X.columns)[
-                             'Probability'])
+    result_actual = list(my_logistic_regresion_class.get_interpretation()['Probability'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (result_required == result_actual)
+
+    result_required = [0.22395117322839098, 0.23089275089703268, 0.35583110026571313,
+                       0.3977261128754439, 0.4027792667251947, 0.4068054774641371]
+    result_actual = sorted(list(my_logistic_regresion_class.feature_interpretability_comparative_df['Prediction_add1_or_with']))
 
     result_required = list(map(lambda x: round(x, 2), result_required))
     result_actual = list(map(lambda x: round(x, 2), result_actual))
@@ -549,9 +682,7 @@ auc = 0.85'''
     assert (result_required == result_actual)
 
     result_required = [0.4061624649859944, 0.236473141666754, 0.7141621577909735, 0.3998399415589254, 0.3537524130019424]
-    result_actual = list(my_logistic_regresion_class.get_interpretation(my_logistic_regresion_class.result_with_feature_selection,
-                                                                        my_logistic_regresion_class.X_with_feature_selection.columns)[
-                             'Probability'])
+    result_actual = list(my_logistic_regresion_class.get_interpretation()['Probability'])
 
     result_required = list(map(lambda x: round(x, 2), result_required))
     result_actual = list(map(lambda x: round(x, 2), result_actual))
@@ -570,6 +701,195 @@ auc = 0.85'''
     assert (result_required == result_actual)
 
 
+    print('Success!')
+
+def unit_test_5():
+    print('Unit test 5...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    df.loc[1,'Survived'] = np.nan
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05)
+    my_logistic_regresion_class.log_reg()
+
+    result_required = [5.38, -1.24, -0.04, -0.38, -0.06, 0.0, -2.63]
+    result_actual = list(my_logistic_regresion_class.result.params)
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
+
+    print('Success!')
+
+def unit_test_6():
+    print('Unit test 6...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    df.loc[1,'Survived'] = np.nan
+    for col in df.columns:
+        if col in ['Pclass','Parch']:
+            df[col] = df[col].astype('str')
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05,cols_to_keep_static=['Pclass'])
+    my_logistic_regresion_class.log_reg_with_feature_selection(verbose=False)
+
+    result_required = [1.7, -1.41, -2.65, 2.62, -0.04, -0.38]
+    result_actual = list(my_logistic_regresion_class.result_with_feature_selection.params)
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
+    print('Success!')
+
+
+def unit_test_7():
+    print('Unit test 7...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    for col in df.columns:
+        if col in ['Pclass','Parch']:
+            df[col] = df[col].astype('str')
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05,cols_to_keep_static=['Pclass'],
+                                                          cols_to_try_individually=['Parch','Sex','Age','Fare'],
+                                                          max_iter=1000)
+    my_logistic_regresion_class.log_reg_one_at_a_time(with_feature_selection=True)
+
+    result_required = [-0.73, 2.64, -0.04, 0.01]
+    result_actual = list(my_logistic_regresion_class.df_one_at_a_time['Coefficient'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
+    print('Success!')
+
+def unit_test_8():
+    print('Unit test 8...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    for col in df.columns:
+        if col in ['Pclass','Parch']:
+            df[col] = df[col].astype('str')
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05,cols_to_keep_static=['Pclass'],
+                                                          cols_to_try_individually=['Age','Fare'],
+                                                          max_iter=1000)
+    my_logistic_regresion_class.log_reg_one_at_a_time(with_feature_selection=False)
+
+    result_required = [-0.04, 0.01]
+    result_actual = list(my_logistic_regresion_class.df_one_at_a_time['Coefficient'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
+    print('Success!')
+
+def unit_test_9():
+    print('Unit test 9...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    for col in df.columns:
+        if col in ['Pclass','Parch']:
+            df[col] = df[col].astype('str')
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05,cols_to_keep_static=['Pclass'],
+                                                          cols_to_try_individually=['Age','Fare'],
+                                                          max_iter=1000)
+    my_logistic_regresion_class.log_reg_one_at_a_time(with_feature_selection=False,get_interpretability=True)
+
+    result_required = [0.4, 0.39]
+    result_actual = list(my_logistic_regresion_class.df_one_at_a_time['Controlled Probability'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
+    print('Success!')
+
+def unit_test_10():
+    print('Unit test 10...')
+    import sys
+    import os
+    import warnings
+
+    np.random.seed(101)
+    #warnings.filterwarnings("ignore")
+
+    current_dir = '/'.join(sys.path[0].split('/')[:-1])  # sys.path[0]
+    data_dir = os.path.join(current_dir, 'Data', 'titanic')
+    titanic_csv = os.path.join(data_dir, 'titanic.csv')
+    df = pd.read_csv(titanic_csv)
+    df = df[['Survived', 'Pclass', 'Sex', 'Age', 'SibSp','Parch', 'Fare']]
+    for col in df.columns:
+        if col in ['Pclass','Parch']:
+            df[col] = df[col].astype('str')
+    my_logistic_regresion_class = LogisticRegressionClass(df,'Survived',sig_level=0.05,cols_to_keep_static=['Pclass'],
+                                                          cols_to_try_individually=['Age','Fare','Parch','Sex'],
+                                                          max_iter=1000)
+    my_logistic_regresion_class.log_reg_one_at_a_time(with_feature_selection=True,get_interpretability=True)
+
+    result_required = [0.4, 0.39, 0.35, 0.72]
+    result_actual = list(my_logistic_regresion_class.df_one_at_a_time['Controlled Probability'])
+
+    result_required = list(map(lambda x: round(x, 2), result_required))
+    result_actual = list(map(lambda x: round(x, 2), result_actual))
+
+    assert (sorted(result_required) == sorted(result_actual))
+
     print('Success!')
 
 if __name__ == '__main__':
@@ -577,3 +897,9 @@ if __name__ == '__main__':
     unit_test_2()
     unit_test_3()
     unit_test_4()
+    unit_test_5()
+    unit_test_6()
+    unit_test_7()
+    unit_test_8()
+    unit_test_9()
+    unit_test_10()
